/* Copyright (C) 2015-2020 |Meso|Star> (contact@meso-star.com)
 *
 * This software is a computer program whose purpose is to describe a
 * virtual 3D environment that can be ray-traced and sampled both robustly
 * and efficiently.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#include "s3d.h"
#include "s3d_device_c.h"
#include "s3d_instance.h"
#include "s3d_geometry.h"
#include "s3d_mesh.h"
#include "s3d_scene_view_c.h"
#include "s3d_sphere.h"

#include <rsys/float3.h>
#include <rsys/float33.h>

struct point_query_context {
  struct RTCPointQueryContext rtc;
  struct s3d_scene_view* scnview;
  void* data; /* Per point query defined data */
};

/*******************************************************************************
 * Helper functions
 ******************************************************************************/
static INLINE float*
closest_point_triangle
  (const float p[3], /* Point */
   const float a[3], /* 1st triangle vertex */
   const float b[3], /* 2nd triangle vertex */
   const float c[3], /* 3rd triangle vertex */
   float closest_pt[3], /* Closest position */
   float uv[2]) /* UV of the closest position */
{
  float ab[3], ac[3], ap[3], bp[3], cp[3];
  float d1, d2, d3, d4, d5, d6;
  float va, vb, vc;
  float rcp_triangle_area;
  float v, w;
  ASSERT(p && a && b && c && closest_pt && uv);

  f3_sub(ab, b, a);
  f3_sub(ac, c, a);

  /* Check if the closest point is the triangle vertex 'a' */
  f3_sub(ap, p, a);
  d1 = f3_dot(ab, ap);
  d2 = f3_dot(ac, ap);
  if(d1 <= 0.f && d2 <= 0.f) {
    uv[0] = 1.f;
    uv[1] = 0.f;
    return f3_set(closest_pt, a);
  }

  /* Check if the closest point is the triangle vertex 'b' */
  f3_sub(bp, p, b);
  d3 = f3_dot(ab, bp);
  d4 = f3_dot(ac, bp);
  if(d3 >= 0.f && d4 <= d3) {
    uv[0] = 0.f;
    uv[1] = 1.f;
    return f3_set(closest_pt, b);
  }

  /* Check if the closest point is the triangle vertex 'c' */
  f3_sub(cp, p, c);
  d5 = f3_dot(ab, cp);
  d6 = f3_dot(ac, cp);
  if(d6 >= 0.f && d5 <= d6) {
    uv[0] = 0.f;
    uv[1] = 0.f;
    return f3_set(closest_pt, c);
  }

  /* Check if the closest point is on the triangle edge 'ab' */
  vc = d1*d4 - d3*d2;
  if(vc <= 0.f && d1 >= 0.f && d3 <= 0.f) {
    const float s = d1 / (d1 - d3);
    closest_pt[0] = a[0] + s*ab[0];
    closest_pt[1] = a[1] + s*ab[1];
    closest_pt[2] = a[2] + s*ab[2];
    uv[0] = 1.f - s;
    uv[1] = s;
    return closest_pt;
  }

  /* Check if the closest point is on the triangle edge 'ac' */
  vb = d5*d2 - d1*d6;
  if(vb <= 0.f && d2 >= 0.f && d6 <= 0.f) {
    const float s = d2 / (d2 - d6);
    closest_pt[0] = a[0] + s*ac[0];
    closest_pt[1] = a[1] + s*ac[1];
    closest_pt[2] = a[2] + s*ac[2];
    uv[0] = 1.f - s;
    uv[1] = 0.f;
    return closest_pt;
  }

  /* Check if the closest point is on the triangle edge 'bc' */
  va = d3*d6 - d5*d4;
  if(va <= 0.f && (d4 - d3) >= 0.f && (d5 - d6) >= 0.f) {
    const float s = (d4 - d3) / ((d4 - d3) + (d5 - d6));
    closest_pt[0] = b[0] + s*(c[0] - b[0]);
    closest_pt[1] = b[1] + s*(c[1] - b[1]);
    closest_pt[2] = b[2] + s*(c[2] - b[2]);
    uv[0] = 0.f;
    uv[1] = 1.f - s;
    return closest_pt;
  }

  /* The closest point lies in the triangle: compute its barycentric
   * coordinates */
  rcp_triangle_area = 1.f / (va + vb + vc);
  v = vb * rcp_triangle_area;
  w = vc * rcp_triangle_area;

  /* Save the uv barycentric coordinates */
  uv[0] = 1.f - v - w;
  uv[1] = v;
  ASSERT(eq_epsf(uv[0] + uv[1] + w, 1.f, 1.e-4f));

  if(uv[0] < 0.f) { /* Handle precision issues */
    if(uv[1] > w) uv[1] += uv[0];
    uv[0] = 0;
  }

  /* Use the barycentric coordinates to compute the world space position of the
   * closest point onto the triangle */
  closest_pt[0] = a[0] + v*ab[0] + w*ac[0];
  closest_pt[1] = a[1] + v*ab[1] + w*ac[1];
  closest_pt[2] = a[2] + v*ab[2] + w*ac[2];
  return closest_pt;
}

static bool
closest_point_mesh
  (struct RTCPointQueryFunctionArguments* args,
   struct geometry* geom,
   struct geometry* inst, /* Can be NULL */
   void* query_data)
{
  struct s3d_hit hit = S3D_HIT_NULL;
  struct s3d_hit* out_hit = NULL;
  struct hit_filter* filter = NULL;
  const uint32_t* ids = NULL;
  float closest_point[3];
  float query_pos_ws[3]; /* World space query position */
  float query_pos_ls[3]; /* Local space query position */
  float v0[3], v1[3], v2[3];
  float E0[3], E1[3], Ng[3];
  float vec[3];
  float uv[2];
  float dst;
  int flip_surface = 0;
  ASSERT(args && geom && geom->type == GEOM_MESH);
  ASSERT(args->primID < mesh_get_ntris(geom->data.mesh));

  /* Fetch triangle indices */
  ids = mesh_get_ids(geom->data.mesh) + args->primID*3/*#indices per triangle*/;

  /* Fetch triangle vertices */
  ASSERT(geom->data.mesh->attribs_type[S3D_POSITION] == S3D_FLOAT3);
  f3_set(v0, mesh_get_pos(geom->data.mesh) + ids[0]*3/*#coords per vertex*/);
  f3_set(v1, mesh_get_pos(geom->data.mesh) + ids[1]*3/*#coords per vertex*/);
  f3_set(v2, mesh_get_pos(geom->data.mesh) + ids[2]*3/*#coords per vertex*/);

  /* Local copy of the query position */
  query_pos_ws[0] = args->query->x;
  query_pos_ws[1] = args->query->y;
  query_pos_ws[2] = args->query->z;

  if(!args->context->instStackSize) { /* The mesh is instantiated */
    query_pos_ls[0] = query_pos_ws[0];
    query_pos_ls[1] = query_pos_ws[1];
    query_pos_ls[2] = query_pos_ws[2];
  } else {
    const float* world2inst;
    float a[3], b[3], c[3];
    ASSERT(args->context->instStackSize == 1);
    ASSERT(inst && inst->type == GEOM_INSTANCE);

    world2inst = args->context->world2inst[0];

    /* Transform the query position in instance space */
    f3_mulf(a, world2inst+0, query_pos_ws[0]);
    f3_mulf(b, world2inst+4, query_pos_ws[1]);
    f3_mulf(c, world2inst+8, query_pos_ws[2]);
    query_pos_ls[0] = a[0] + b[0] + c[0] + world2inst[12];
    query_pos_ls[1] = a[1] + b[1] + c[1] + world2inst[13];
    query_pos_ls[2] = a[2] + b[2] + c[2] + world2inst[14];

    flip_surface = inst->flip_surface;
  }

  /* Compute the closest point onto the triangle from the submitted point */
  closest_point_triangle(query_pos_ls, v0, v1, v2, closest_point, uv);

  /* Compute the distance */
  f3_sub(vec, closest_point, query_pos_ls);
  dst = f3_len(vec);

  /* Transform the distance in world space */
  if(args->context->instStackSize != 0) {
    ASSERT(args->similarityScale > 0);
    dst /= args->similarityScale;
  }

  if(dst >= args->query->radius) return 0;

  /* Compute the triangle normal in world space */
  f3_sub(E0, v2, v0);
  f3_sub(E1, v1, v0);
  f3_cross(Ng, E0, E1);

  /* Flip the geometric normal wrt the flip surface flag */
  flip_surface ^= geom->flip_surface;
  if(flip_surface) f3_minus(Ng, Ng);

  /* Setup the hit */
  hit.prim.shape__ = geom;
  hit.prim.inst__ = inst;
  hit.distance = dst;
  hit.uv[0] = uv[0];
  hit.uv[1] = uv[1];
  hit.normal[0] = Ng[0];
  hit.normal[1] = Ng[1];
  hit.normal[2] = Ng[2];
  hit.prim.prim_id = args->primID;
  hit.prim.geom_id = geom->name;
  hit.prim.inst_id = inst ? inst->name : S3D_INVALID_ID;
  hit.prim.scene_prim_id =
    hit.prim.prim_id
  + geom->scene_prim_id_offset
  + (inst ? inst->scene_prim_id_offset : 0);

  /* `vec' is the direction along which the closest point was found. We thus
   * submit it to the filter function as the direction corresponding to the
   * computed hit */
  filter = &geom->data.mesh->filter;
  if(filter->func
  && filter->func(&hit, query_pos_ws, vec, query_data, filter->data)) {
    return 0; /* This point is filtered. Discard it! */
  }

  /* Update output data */
  out_hit = args->userPtr;
  *out_hit = hit;

  /* Shrink the query radius */
  args->query->radius = dst;

  return 1; /* Notify that the query radius was updated */
}

static bool
closest_point_sphere
  (struct RTCPointQueryFunctionArguments* args,
   struct geometry* geom,
   struct geometry* inst,
   void* query_data)
{
  struct s3d_hit hit = S3D_HIT_NULL;
  struct s3d_hit* out_hit = NULL;
  struct hit_filter* filter = NULL;
  float query_pos[3];
  float sphere_pos[3];
  float Ng[3];
  float dir[3];
  float uv[2];
  float dst;
  float len;
  int flip_surface = 0;
  ASSERT(args && geom && geom->type == GEOM_SPHERE && args->primID == 0);

  /* Local copy of the query position */
  query_pos[0] = args->query->x;
  query_pos[1] = args->query->y;
  query_pos[2] = args->query->z;

  f3_set(sphere_pos, geom->data.sphere->pos);
  if(args->context->instStackSize) { /* The sphere is instantiated */
    const float* transform;
    transform = inst->data.instance->transform;
    ASSERT(args->context->instStackSize == 1);
    ASSERT(inst && inst->type == GEOM_INSTANCE);
    ASSERT(f3_eq_eps(transform+0, args->context->inst2world[0]+0, 1.e-6f));
    ASSERT(f3_eq_eps(transform+3, args->context->inst2world[0]+4, 1.e-6f));
    ASSERT(f3_eq_eps(transform+6, args->context->inst2world[0]+8, 1.e-6f));
    ASSERT(f3_eq_eps(transform+9, args->context->inst2world[0]+12,1.e-6f));

    /* Transform the sphere position in world space */
    f33_mulf3(sphere_pos, transform, sphere_pos);
    f3_add(sphere_pos, transform+9, sphere_pos);

    flip_surface = inst->flip_surface;
  }

  /* Compute the distance from the query pos to the sphere center */
  f3_sub(Ng, query_pos, sphere_pos);
  len = f3_len(Ng);

  /* Evaluate the distance from the query pos to the sphere surface */
  dst = fabsf(len - geom->data.sphere->radius);

  /* The closest point onto the sphere is outside the query radius */
  if(dst >= args->query->radius)
    return 0;

  /* Normalize the hit normal */
  if(len > 0) {
    f3_divf(Ng, Ng, len);
  } else {
    /* The query position is equal to the sphere center. Arbitrarily choose the
     * point along the +X axis as the closest point */
    Ng[0] = 1.f;
    Ng[1] = 0.f;
    Ng[2] = 0.f;
  }

  /* Compute the uv of the found point */
  sphere_normal_to_uv(Ng, uv);

  /* Flip the geometric normal wrt the flip surface flag */
  flip_surface ^= geom->flip_surface;
  if(flip_surface) f3_minus(Ng, Ng);

  /* Setup the hit */
  hit.prim.shape__ = geom;
  hit.prim.inst__ = inst;
  hit.distance = dst;
  hit.uv[0] = uv[0];
  hit.uv[1] = uv[1];
  hit.normal[0] = Ng[0];
  hit.normal[1] = Ng[1];
  hit.normal[2] = Ng[2];
  hit.prim.prim_id = args->primID;
  hit.prim.geom_id = geom->name;
  hit.prim.inst_id = inst ? inst->name : S3D_INVALID_ID;
  hit.prim.scene_prim_id =
    hit.prim.prim_id
  + geom->scene_prim_id_offset
  + (inst ? inst->scene_prim_id_offset : 0);

  /* Use the reversed geometric normal as the hit direction since it is along
   * this vector that the closest point was effectively computed */
  f3_minus(dir, Ng);
  filter = &geom->data.sphere->filter;
  if(filter->func
  && filter->func(&hit, query_pos, dir, query_data, filter->data)) {
    return 0;
  }

  /* Update output data */
  out_hit = args->userPtr;
  *out_hit = hit;

  /* Shrink the query radius */
  args->query->radius = dst;

  return 1; /* Notify that the query radius was updated */
}

static bool
closest_point(struct RTCPointQueryFunctionArguments* args)
{
  struct point_query_context* ctx = NULL;
  struct geometry* geom = NULL;
  struct geometry* inst = NULL;
  bool query_radius_is_upd = false;
  ASSERT(args);

  ctx = CONTAINER_OF(args->context, struct point_query_context, rtc);
  if(args->context->instStackSize == 0) {
    geom = scene_view_geometry_from_embree_id
      (ctx->scnview, args->geomID);
  } else {
    ASSERT(args->context->instStackSize == 1);
    inst = scene_view_geometry_from_embree_id
      (ctx->scnview, args->context->instID[0]);
    geom = scene_view_geometry_from_embree_id
      (inst->data.instance->scnview, args->geomID);
  }

  switch(geom->type) {
    case GEOM_MESH:
      query_radius_is_upd = closest_point_mesh(args, geom, inst, ctx->data);
      break;
    case GEOM_SPHERE:
      query_radius_is_upd = closest_point_sphere(args, geom, inst, ctx->data);
      break;
    default: FATAL("Unreachable code\n"); break;
  }
  return query_radius_is_upd;
}

/*******************************************************************************
 * Exported functions
 ******************************************************************************/
res_T
s3d_scene_view_closest_point
  (struct s3d_scene_view* scnview,
   const float pos[3],
   const float radius,
   void* query_data,
   struct s3d_hit* hit)
{
  struct RTCPointQuery query;
  struct point_query_context query_ctx;

  if(!scnview || !pos || radius <= 0 || !hit)
    return RES_BAD_ARG;
  if((scnview->mask & S3D_TRACE) == 0) {
    log_error(scnview->scn->dev,
      "%s: the S3D_TRACE flag is not active onto the submitted scene view.\n",
      FUNC_NAME);
    return RES_BAD_OP;
  }

  *hit = S3D_HIT_NULL;

  /* Initialise the point query */
  query.x = pos[0];
  query.y = pos[1];
  query.z = pos[2];
  query.radius = radius;
  query.time = FLT_MAX; /* Invalid fields */

  /* Initialise the point query context */
  rtcInitPointQueryContext(&query_ctx.rtc);
  query_ctx.scnview = scnview;
  query_ctx.data = query_data;

  /* Here we go! */
  rtcPointQuery(scnview->rtc_scn, &query, &query_ctx.rtc, closest_point, hit);

  return RES_OK;
}

