/* Copyright (C) 2015-2020 |Meso|Star> (contact@meso-star.com)
 *
 * This software is a computer program whose purpose is to describe a
 * virtual 3D environment that can be ray-traced and sampled both robustly
 * and efficiently.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#ifndef S3D_SPHERE_H
#define S3D_SPHERE_H

#include "s3d_c.h"

#include <rsys/ref_count.h>
#include <rsys/float3.h>

struct sphere {
  float pos[3];
  float radius;
  struct s3d_device* dev;
  struct hit_filter filter;
  ref_T ref;
};

extern LOCAL_SYM res_T
sphere_create
  (struct s3d_device* dev,
   struct sphere** sphere);

extern LOCAL_SYM void
sphere_ref_get
  (struct sphere* sphere);

extern LOCAL_SYM void
sphere_ref_put
  (struct sphere* sphere);

static INLINE int
sphere_is_degenerated(const struct sphere* sphere)
{
  ASSERT(sphere);
  return sphere->radius < 0;
}

static INLINE void
sphere_compute_aabb
  (const struct sphere* sphere,
   float lower[3],
   float upper[3])
{
  ASSERT(sphere && lower && upper);
  if(sphere_is_degenerated(sphere)) {
    f3_splat(lower, FLT_MAX);
    f3_splat(upper,-FLT_MAX);
  } else {
    f3_subf(lower, sphere->pos, sphere->radius);
    f3_addf(upper, sphere->pos, sphere->radius);
  }
}

static INLINE float
sphere_compute_area(const struct sphere* sphere)
{
  float r;
  ASSERT(sphere);
  r = sphere->radius;
  return (float)(4*PI*r*r);
}

static INLINE float
sphere_compute_volume(const struct sphere* sphere)
{
  float r;
  ASSERT(sphere);
  r = sphere->radius;
  return (float)(4*PI*r*r*r/3);
}

static FINLINE void
sphere_normal_to_uv(const float normal[3], float uv[2])
{
  float u, v, cos_theta;
  ASSERT(normal && uv && f3_is_normalized(normal));

  cos_theta = normal[2];

  v = (1.f - cos_theta) * 0.5f;
  if(absf(cos_theta) == 1) {
    u = 0;
  } else if(eq_epsf(normal[0], 0.f, 1.e-6f)) {
    u = normal[1] > 0 ? 0.25f : 0.75f;
  } else {
    double phi = atan2f(normal[1], normal[0]); /* phi in [-PI, PI] */
    if(phi < 0) phi = 2*PI + phi; /* phi in [0, 2PI] */
    u = (float)(phi / (2*PI));
  }
  uv[0] = u;
  uv[1] = v;
}

#endif /* S3D_SPHERE_H */
