/* Copyright (C) |Meso|Star> 2015-2016 (contact@meso-star.com)
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#define _POSIX_C_SOURCE 200112L /* strtok_r support */

#include "sstl.h"

#include <rsys/cstr.h>
#include <rsys/float3.h>
#include <rsys/hash_table.h>
#include <rsys/logger.h>
#include <rsys/mem_allocator.h>
#include <rsys/ref_count.h>
#include <rsys/stretchy_array.h>

#include <string.h>

#ifdef COMPILER_CL
  #pragma warning(push)
  #pragma warning(disable:4706) /* Assignment within a condition */

  #define strtok_r strtok_s
#endif

struct solid {
  char* name;
  unsigned* indices;
  float* vertices;
  float* normals;
};
static const struct solid SOLID_NULL = { NULL, NULL, NULL, NULL };

struct streamer {
  FILE* stream;
  const char* name;
  size_t iline;
  char* buf;
};

struct vertex { float xyz[3]; };

static INLINE char
eq_vertex(const struct vertex* a, const struct vertex* b)
{
  return (char)f3_eq(a->xyz, b->xyz);
}

/* Declare the hash table that map a vertex to its index */
#define HTABLE_NAME vertex
#define HTABLE_DATA unsigned
#define HTABLE_KEY struct vertex
#define HTABLE_KEY_FUNCTOR_EQ eq_vertex
#include <rsys/hash_table.h>

struct sstl {
  int verbose;
  struct htable_vertex vertex2id;
  struct solid solid;

  struct logger* logger;
  struct mem_allocator* allocator;
  ref_T ref;
};

/*******************************************************************************
 * Helper functions
 ******************************************************************************/
static void
streamer_init(struct streamer* streamer, FILE* stream, const char* name)
{
  ASSERT(streamer && stream && name);
  memset(streamer, 0, sizeof(struct streamer));
  streamer->stream = stream;
  streamer->name = name;
  streamer->iline = 0;
  streamer->buf = sa_add(streamer->buf, 128);
  ASSERT(streamer->buf);
}

static void
streamer_release(struct streamer* streamer)
{
  ASSERT(streamer);
  sa_release(streamer->buf);
}

static char*
streamer_read_line(struct streamer* streamer)
{
  const size_t buf_chunk = 256;
  char* line;
  ASSERT(streamer);

  for(;(line=fgets(streamer->buf, (int)sa_size(streamer->buf), streamer->stream));
      ++streamer->iline) {

    /* Ensure that the whole line is read */
    while(!strrchr(line, '\n')) {
      char* remain = sa_add(streamer->buf, buf_chunk);
      if(!remain) {
        FATAL("Not enough memory: couldn't resize the stream buffer.\n");
      }
      line = streamer->buf;
      if(!fgets(remain, (int)buf_chunk, streamer->stream)) /* EOF */
        break;
    }

    if(strspn(streamer->buf, " \t\r\n") != strlen(streamer->buf)) { /* Not empty */
      /* Remove newline character(s) */
      size_t last_char = strlen(line);
      while(last_char-- && (line[last_char]=='\n' || line[last_char]=='\r'));
      line[last_char + 1] = '\0';
      break;
    }
  }
  ++streamer->iline;
  return line;
}

static void
solid_clear(struct solid* solid)
{
  ASSERT(solid);
  sa_release(solid->name);
  sa_release(solid->vertices);
  sa_release(solid->indices);
  sa_release(solid->normals);
}

static void
clear(struct sstl* sstl)
{
  ASSERT(sstl);
  solid_clear(&sstl->solid);
  sstl->solid = SOLID_NULL;
  htable_vertex_clear(&sstl->vertex2id);
}

static void
print_log
  (const struct sstl* sstl, const enum log_type type, const char* msg, ...)
{
  va_list vargs_list;
  ASSERT(sstl && msg);
  if(sstl->verbose) {
    res_T res; (void)res;
    va_start(vargs_list, msg);
    res = logger_vprint(sstl->logger, type, msg, vargs_list);
    ASSERT(res == RES_OK);
    va_end(vargs_list);
  }
}

static INLINE res_T
parse_float3
  (struct sstl* sstl,
   char* str,
   float vert[3],
   const char* filename,
   const size_t iline,
   char** tok_ctx)
{
  char* tk;
  int i;
  res_T res = RES_OK;
  ASSERT(str && vert && filename);

  FOR_EACH(i, 0, 3) {
    tk = strtok_r(i==0 ? str : NULL, " \t", tok_ctx);
    if(!tk) {
      print_log(sstl, LOG_ERROR, "%s:%lu: expecting 3D coordinates.\n",
        filename, (unsigned long)iline);
      return RES_BAD_ARG;
    }

    res = cstr_to_float(tk, vert + i);
    if(res != RES_OK) {
      print_log(sstl, LOG_ERROR, "%s:%lu: invalid coordinate \"%s\".\n",
        filename, (unsigned long)iline, tk);
      return res;
    }
  }
  tk = strtok_r(NULL, "\0", tok_ctx);
  if(tk) { /* Unexpected remaining chars */
    print_log(sstl, LOG_WARNING, "%s:%lu: unexpected directive \"%s\".\n",
      filename, (unsigned long)iline, tk);
  }
  return RES_OK;
}

static INLINE res_T
parse_name_string
  (struct sstl* sstl,
   char* str,
   char** out_name,
   const char* filename,
   const size_t iline,
   char** tok_ctx)
{
  char* name = NULL;
  char* tk;
  res_T res = RES_OK;
  ASSERT(sstl && out_name);

  if(!str) goto exit;

  /* Handle name with spaces */
  for(tk = strtok_r(str, " \t", tok_ctx); tk; tk = strtok_r(NULL, " \t", tok_ctx)) {
    char* remain = NULL;
    if(name) name[strlen(name)] = ' '; /* Replace '\0' by ' ' */
    remain = sa_add(name, strlen(tk) + 1/*NULL char*/);
    if(!remain) {
      print_log(sstl, LOG_ERROR,
        "%s:%lu: not enough memory: couldn't allocate the name of the solid.\n",
        filename, (unsigned long)iline);
      res = RES_MEM_ERR;
      goto error;
    }
    strcpy(name, tk);
  }

exit:
  *out_name = name;
  return res;
error:
  if(name) {
    sa_release(name);
    name = NULL;
  }
  goto exit;
}

static INLINE res_T
parse_solid_name
  (struct sstl* sstl,
   struct solid* solid,
   char* line,
   const char* filename,
   const size_t iline,
   char** tok_ctx)
{
  res_T res = RES_OK;
  ASSERT(sstl && solid && !solid->name);

  if(!line || strcmp(strtok_r(line, " \t", tok_ctx), "solid")) {
    print_log(sstl, LOG_ERROR,
      "%s:%lu: missing the \"solid [NAME]\" directive.\n",
      filename, (unsigned long)iline);
    res = RES_BAD_ARG;
    goto error;
  }

  res = parse_name_string
    (sstl, strtok_r(NULL, "\0", tok_ctx), &solid->name, filename, iline, tok_ctx);
  if(res != RES_OK) goto error;

exit:
  return res;
error:
  if(solid->name) sa_release(solid->name);
  goto exit;
}

static INLINE res_T
parse_solid_vertex
  (struct sstl* sstl,
   struct solid* solid,
   unsigned* const index,
   char* line,
   const char* filename,
   const size_t iline,
   char** tok_ctx)
{
  struct vertex vertex;
  unsigned* found_id;
  res_T res = RES_OK;
  ASSERT(sstl && solid && index);

  if(!line || strcmp(strtok_r(line, " \t", tok_ctx), "vertex")) {
    print_log(sstl, LOG_ERROR,
      "%s:%lu: missing a \"vertex X Y Z\" directive.\n",
      filename, (unsigned long)iline);
    return RES_BAD_ARG;
  }

  res = parse_float3(sstl, strtok_r(NULL, "\0", tok_ctx), vertex.xyz, filename, iline, tok_ctx);
  if(res != RES_OK) return res;

  /* Look for an already registered vertex position */
  found_id = htable_vertex_find(&sstl->vertex2id, &vertex);

  if(found_id) {
    *index = *found_id;
  } else {
    /* Add a new vertex */
    *index = (unsigned)sa_size(solid->vertices) / 3;
    res = htable_vertex_set(&sstl->vertex2id, &vertex, index);
    if(res != RES_OK) {
      print_log(sstl, LOG_ERROR,
        "%s:%lu: couldn't register a vertex position.\n",
        filename, (unsigned long)iline);
    }
    f3_set(sa_add(solid->vertices, 3), vertex.xyz);
  }
  sa_push(solid->indices, *index);

  return RES_OK;
}

static INLINE res_T
parse_outer_loop
  (struct sstl* sstl,
   char* line,
   const char* filename,
   const size_t iline,
   char** tok_ctx)
{
  char* tk;
  ASSERT(sstl);

  if(!line
  || strcmp(strtok_r(line, " \t", tok_ctx), "outer")
  || !(tk = strtok_r(NULL, " \t", tok_ctx))
  || strcmp(tk, "loop")) {
    print_log(sstl, LOG_ERROR,
      "%s:%lu: missing the \"outer loop\" directive.\n",
      filename, (unsigned long)iline);
    return RES_BAD_ARG;
  }

  tk = strtok_r(NULL, "\0", tok_ctx);
  if(tk && strspn(tk, " \t\r\n") != strlen(tk)) { /* Invalid remaining chars */
    print_log(sstl, LOG_WARNING,
      "%s:%lu: malformed \"outer loop\" directive.\n",
      filename, (unsigned long)iline);
  }
  return RES_OK;
}

static INLINE res_T
parse_directive
  (struct sstl* sstl,
   const char* directive,
   char* line,
   const char* filename,
   const size_t iline,
   char** tok_ctx)
{
  char* tk;
  ASSERT(sstl && directive);

  if(!line || strcmp(strtok_r(line, " \t", tok_ctx), directive)) {
    print_log(sstl, LOG_ERROR,
      "%s:%lu: missing the \"%s\" directive.\n",
      filename, (unsigned long)iline, directive);
    return RES_BAD_ARG;
  }

  tk = strtok_r(NULL, " \0", tok_ctx);
  if(tk && strspn(tk, " \t\r\n") != strlen(tk)) { /* Invalid remaining chars */
    print_log(sstl, LOG_WARNING,
      "%s:%lu: malformed \"%s\" directive.\n",
      filename, (unsigned long)iline, directive);
  }

  return RES_OK;
}

static res_T
load_stream
  (struct sstl* sstl,
   FILE* stream,
   const char* stream_name,
   char** tok_ctx)
{
  res_T res = RES_OK;
  struct streamer streamer;
  struct solid solid = SOLID_NULL;
  char* line = NULL;
  char* tk;

  ASSERT(sstl && stream);
  streamer_init(&streamer, stream, stream_name);
  clear(sstl);

  line = streamer_read_line(&streamer);
  res = parse_solid_name(sstl, &solid, line, streamer.name, streamer.iline, tok_ctx);
  if(res != RES_OK) goto error;

  for(;;) { /* Parse the solid facets */
    float normal[3];
    unsigned facet[3];
    int ivertex;

    line = streamer_read_line(&streamer);
    if(!line) {
      print_log(sstl, LOG_ERROR, "%s:%lu: missing directive.\n",
        streamer.name, (unsigned long)streamer.iline);
      res = RES_BAD_ARG;
      goto error;
    }

    tk = strtok_r(line, " \t", tok_ctx);

    /* Stop on "endsolid" directive */
    if(!strcmp(tk, "endsolid"))
      break;

    /* Parse the facet normal directive */
    if(strcmp(tk, "facet")
    || !(tk = strtok_r(NULL, " \t", tok_ctx))
    || strcmp(tk, "normal")) {
      print_log(sstl, LOG_ERROR,
        "%s:%lu: missing or malformed \"facet normal X Y Z\" directive.\n",
        streamer.name, (unsigned long)streamer.iline);
      res = RES_BAD_ARG;
      goto error;
    }
    res = parse_float3
      (sstl, strtok_r(NULL, "\0", tok_ctx), normal, streamer.name, streamer.iline, tok_ctx);
    if(res != RES_OK) goto error;

    /* Parse the Outer loop directive */
    line = streamer_read_line(&streamer);
    res = parse_outer_loop(sstl, line, streamer.name, streamer.iline, tok_ctx);
    if(res != RES_OK) goto error;

    /* Parse the facet vertices. Assume that only 3 vertices are submitted */
    FOR_EACH(ivertex, 0, 3) {
      line = streamer_read_line(&streamer);
      res = parse_solid_vertex
        (sstl, &solid, facet+ivertex, line, streamer.name, streamer.iline, tok_ctx);
      if(res != RES_OK) goto error;
    }

    if(!f3_is_normalized(normal)) { /* Use geometry normal */
      float v0[3], v1[3];
      /* Vertices are CCW ordered and the normal follows the right handed rule */
      f3_sub(v0, solid.vertices + facet[1]*3, solid.vertices + facet[0]*3);
      f3_sub(v1, solid.vertices + facet[2]*3, solid.vertices + facet[0]*3);
      f3_cross(normal, v0, v1);
      f3_normalize(normal, normal);
    }
    f3_set(sa_add(solid.normals, 3), normal);

    line = streamer_read_line(&streamer);
    res = parse_directive(sstl, "endloop", line, streamer.name, streamer.iline, tok_ctx);
    if(res != RES_OK) goto error;

    line = streamer_read_line(&streamer);
    res = parse_directive(sstl, "endfacet", line, streamer.name, streamer.iline, tok_ctx);
    if(res != RES_OK) goto error;
  }

  /* Check the solid/endsolid name consistency */
  if(sstl->verbose && solid.name) {
    char* name = NULL;
    res = parse_name_string
      (sstl, strtok_r(NULL, "\0", tok_ctx), &name, streamer.name, streamer.iline, tok_ctx);
    if(res != RES_OK) goto error;

    /* Compare the "endsolid" name with the one of the "solid" directive */
    if(name && strcmp(name, solid.name)) {
      print_log(sstl, LOG_WARNING,
        "%s:%lu: inconsistent \"endsolid\" name.\n",
        streamer.name, (unsigned long)streamer.iline);
    }
    sa_release(name);
  }

  /* Register the solid */
  sstl->solid = solid;

exit:
  streamer_release(&streamer);
  return res;
error:
  solid_clear(&solid);
  goto exit;
}

static void
sstl_release(ref_T* ref)
{
  struct sstl* sstl;
  ASSERT(ref);
  sstl = CONTAINER_OF(ref, struct sstl, ref);
  clear(sstl);
  htable_vertex_release(&sstl->vertex2id);
  MEM_RM(sstl->allocator, sstl);
}

/*******************************************************************************
 * Exported functions
 ******************************************************************************/
res_T
sstl_create
  (struct logger* log,
   struct mem_allocator* mem_allocator,
   const int verbose,
   struct sstl** out_sstl)
{
  struct mem_allocator* allocator = NULL;
  struct logger* logger = NULL;
  struct sstl* sstl = NULL;
  res_T res = RES_OK;

  if(!out_sstl) {
    res = RES_BAD_ARG;
    goto error;
  }

  allocator = mem_allocator ? mem_allocator : &mem_default_allocator;
  logger = log ? log : LOGGER_DEFAULT;

  sstl = MEM_CALLOC(allocator, 1, sizeof(struct sstl));
  if(!sstl) {
    if(verbose) {
      logger_print(logger, LOG_ERROR,
        "Couldn't allocate the Star-STL device.\n");
    }
    res = RES_MEM_ERR;
    goto error;
  }

  ref_init(&sstl->ref);
  sstl->allocator = allocator;
  sstl->logger = logger;
  sstl->verbose = verbose;
  htable_vertex_init(allocator, &sstl->vertex2id);

exit:
  if(out_sstl) *out_sstl = sstl;
  return res;
error:
  if(sstl) {
    SSTL(ref_put(sstl));
    sstl = NULL;
  }
  goto exit;
}

res_T
sstl_ref_get(struct sstl* sstl)
{
  if(!sstl) return RES_BAD_ARG;
  ref_get(&sstl->ref);
  return RES_OK;
}

res_T
sstl_ref_put(struct sstl* sstl)
{
  if(!sstl) return RES_BAD_ARG;
  ref_put(&sstl->ref, sstl_release);
  return RES_OK;

}

res_T
sstl_load(struct sstl* sstl, const char* filename)
{
  FILE* file;
  char* tok_ctx;
  res_T res = RES_OK;

  if(!sstl || !filename)
    return RES_BAD_ARG;

  file = fopen(filename, "r");
  if(!file) {
    print_log(sstl, LOG_ERROR, "Error opening `%s'.\n", filename);
    return RES_IO_ERR;
  }
  res = load_stream(sstl, file, filename, &tok_ctx);
  fclose(file);
  return res;
}

res_T
sstl_load_stream(struct sstl* sstl, FILE* stream)
{
  char* tok_ctx;
  if(!sstl || !stream) return RES_BAD_ARG;
  return load_stream(sstl, stream, "STREAM", &tok_ctx);
}

res_T
sstl_get_desc(const struct sstl* sstl, struct sstl_desc* desc)
{
  if(!sstl || !desc) return RES_BAD_ARG;
  desc->solid_name = sstl->solid.name;
  desc->vertices_count = sa_size(sstl->solid.vertices);
  ASSERT(desc->vertices_count % 3 == 0);
  desc->vertices_count /= 3/* # coords per vertex */;
  desc->triangles_count = sa_size(sstl->solid.indices);
  ASSERT(desc->triangles_count % 3 == 0);
  desc->triangles_count /= 3/* # indices per triange */;
  desc->vertices = sstl->solid.vertices;
  desc->indices = sstl->solid.indices;
  desc->normals = sstl->solid.normals;
  return RES_OK;
}

#ifdef COMPILER_CL
  #pragma warning(pop)
#endif

