/* Copyright (C) 2016-2020 |Meso|Star> (contact@meso-star.com)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>. */

#ifndef SDIS_SCENE_DIMENSION

#ifndef SDIS_SCENE_XD_H
#define SDIS_SCENE_XD_H

#include "sdis_interface_c.h"
#include "sdis_log.h"
#include "sdis_medium_c.h"
#include "sdis_scene_c.h"

#include <star/ssp.h>
#include <rsys/cstr.h>
#include <rsys/float22.h>
#include <rsys/float33.h>
#include <rsys/rsys.h>

/* Emperical cos threshold defining if an angle is sharp */
#define SHARP_ANGLE_COS_THRESOLD -0.70710678 /* ~ cos(3*PI/4) */

/*******************************************************************************
 * Define the helper functions and the data types used by the scene
 * independently of its dimension, i.e. 2D or 3D.
 ******************************************************************************/
/* Context used to wrap the user geometry and interfaces to Star-Enc */
struct geometry {
  void (*indices)(const size_t iprim, size_t ids[], void*);
  void (*interf)(const size_t iprim, struct sdis_interface**, void*);
  void (*position)(const size_t ivert, double pos[], void*);
  void* data;
};

/* Fetch the media split by the primitive `iprim'. This first and second media
 * are the media from the front face side and back face side of the primitive,
 * respectively. */
static void
geometry_media(const unsigned iprim, unsigned media[2], void* data)
{
  struct geometry* ctx = data;
  struct sdis_interface* interf;
  ASSERT(ctx && media);
  ctx->interf(iprim, &interf, ctx->data);
  media[0] = medium_get_id(interf->medium_front);
  media[1] = medium_get_id(interf->medium_back);
}

/* Register the submitted medium against the scene if it is not already
 * registered. On registration, no reference is taken onto the medium; the
 * scene references its media through its interfaces and it is thus useless to
 * take another reference onto them. */
static res_T
register_medium(struct sdis_scene* scn, struct sdis_medium* mdm)
{
  unsigned id;
  size_t nmedia;
  res_T res = RES_OK;
  ASSERT(scn && mdm);

  /* Check that the medium is already registered against the scene */
  id = medium_get_id(mdm);
  nmedia = darray_medium_size_get(&scn->media);
  if(id >= nmedia) {
    res = darray_medium_resize(&scn->media, id + 1);
    if(res != RES_OK) return res;
  }
  if(darray_medium_cdata_get(&scn->media)[id]) {
    ASSERT(darray_medium_cdata_get(&scn->media)[id] == mdm);
  } else {
    /* Do not take a reference onto the medium since we already take a
     * reference onto at least one interface that uses it, and thus that has a
     * reference onto it */
    darray_medium_data_get(&scn->media)[id] = mdm;
  }
  return RES_OK;
}

/* Release the reference toward the interfaces and thus clear the list of scene
 * interfaces, the list of scene media, and the list of per-primitive
 * interface. */
static void
clear_properties(struct sdis_scene* scn)
{
  size_t i;
  ASSERT(scn);
  FOR_EACH(i, 0, darray_interf_size_get(&scn->interfaces)) {
    if(darray_interf_cdata_get(&scn->interfaces)[i]) {
      SDIS(interface_ref_put(darray_interf_data_get(&scn->interfaces)[i]));
    }
  }
  darray_interf_clear(&scn->interfaces);
  darray_medium_clear(&scn->media);
  darray_prim_prop_clear(&scn->prim_props);
}

static INLINE int
check_sdis_scene_create_args(const struct sdis_scene_create_args* args)
{
  return args
      && args->get_indices
      && args->get_interface
      && args->get_position
      && args->nprimitives
      && args->nprimitives < UINT_MAX
      && args->nvertices
      && args->nvertices < UINT_MAX
      && args->fp_to_meter > 0;
}

#endif /* SDIS_SCENE_XD_H */
#else /* !SDIS_SCENE_DIMENSION */

#include "sdis_device_c.h"

#include <rsys/float2.h>
#include <rsys/float3.h>
#include <rsys/double2.h>
#include <rsys/double3.h>
#include <rsys/mem_allocator.h>

#include <limits.h>

/* Check the submitted dimension and include its specific headers */
#define SENCXD_DIM SDIS_SCENE_DIMENSION
#if (SDIS_SCENE_DIMENSION == 2)
  #include <star/sencX2d.h>
  #include <star/s2d.h>
#elif (SDIS_SCENE_DIMENSION == 3)
  #include <star/sencX3d.h>
  #include <star/s3d.h>
#else
  #error "Invalid SDIS_SCENE_DIMENSION value."
#endif

/* Syntactic sugar */
#define DIM SDIS_SCENE_DIMENSION

/* Star-XD macros generic to SDIS_SCENE_DIMENSION */
#define sXd(Name) CONCAT(CONCAT(CONCAT(s, DIM), d_), Name)
#define SXD CONCAT(CONCAT(S, DIM), D)
#define SXD_VERTEX_DATA_NULL CONCAT(CONCAT(S,DIM),D_VERTEX_DATA_NULL)
#define SXD_POSITION CONCAT(CONCAT(S, DIM), D_POSITION)
#define SXD_TRACE CONCAT(CONCAT(S,DIM), D_TRACE)
#define SXD_SAMPLE CONCAT(CONCAT(S,DIM), D_SAMPLE)
#define SXD_GET_PRIMITIVE CONCAT(CONCAT(S,DIM), D_GET_PRIMITIVE)
#define SXD_HIT_NONE CONCAT(CONCAT(S,DIM), D_HIT_NONE)
#define SXD_PRIMITIVE_EQ CONCAT(CONCAT(S,DIM), D_PRIMITIVE_EQ)
#define SXD_FLOATX CONCAT(CONCAT(CONCAT(S,DIM), D_FLOAT), DIM)

/* Vector macros generic to SDIS_SCENE_DIMENSION */
#define fX(Func) CONCAT(CONCAT(CONCAT(f, DIM), _), Func)
#define fX_set_dX CONCAT(CONCAT(CONCAT(f, DIM), _set_d), DIM)
#define fXX_mulfX CONCAT(CONCAT(CONCAT(CONCAT(f, DIM), DIM), _mulf), DIM)

/* Macro making generic its subimitted name to SDIS_SCENE_DIMENSION */
#define XD(Name) CONCAT(CONCAT(CONCAT(Name, _), DIM), d)

#if DIM == 2
  #define HIT_ON_BOUNDARY hit_on_vertex
#else
  #define HIT_ON_BOUNDARY hit_on_edge
#endif

/*******************************************************************************
 * Helper functions
 ******************************************************************************/
#if DIM == 2
#define ON_VERTEX_EPSILON 1.e-4f
/* Check that `hit' roughly lies on a vertex. */
static INLINE int
hit_on_vertex
  (const struct s2d_hit* hit,
   const float org[3],
   const float dir[3])
{
  struct s2d_attrib v0, v1;
  float E[2];
  float hit_pos[2];
  float segment_len;
  float hit_len0;
  float hit_len1;
  ASSERT(hit && !S2D_HIT_NONE(hit) && org && dir);

  /* Rertieve the segment vertices */
  S2D(segment_get_vertex_attrib(&hit->prim, 0, S2D_POSITION, &v0));
  S2D(segment_get_vertex_attrib(&hit->prim, 1, S2D_POSITION, &v1));

  /* Compute the length of the segment */
  segment_len = f2_len(f2_sub(E, v1.value, v0.value));

  /* Compute the hit position onto the segment */
  f2_add(hit_pos, org, f2_mulf(hit_pos, dir, hit->distance));

  /* Compute the length from hit position to segment vertices */
  hit_len0 = f2_len(f2_sub(E, v0.value, hit_pos));
  hit_len1 = f2_len(f2_sub(E, v1.value, hit_pos));

  if(hit_len0 / segment_len < ON_VERTEX_EPSILON
  || hit_len1 / segment_len < ON_VERTEX_EPSILON)
    return 1;
  return 0;
}

static int
hit_shared_vertex
  (const struct s2d_primitive* seg0,
   const struct s2d_primitive* seg1,
   const float pos0[2], /* Tested position onto the segment 0 */
   const float pos1[2]) /* Tested Position onto the segment 1 */
{
  struct s2d_attrib seg0_vertices[2]; /* Vertex positions of the segment 0 */
  struct s2d_attrib seg1_vertices[2]; /* Vertex positions of the segment 1 */
  float d0[2], d1[2]; /* temporary vector */
  float seg0_len, seg1_len; /* Length of the segments */
  float tmp0_len, tmp1_len;
  float cos_normals;
  int seg0_vert = -1; /* Id of the shared vertex for the segment 0 */
  int seg1_vert = -1; /* Id of the shared vertex for the segment 1 */
  int seg0_ivertex, seg1_ivertex;
  ASSERT(seg0 && seg1 && pos0 && pos1);

  /* Fetch the vertices of the segment 0 */
  S2D(segment_get_vertex_attrib(seg0, 0, S2D_POSITION, &seg0_vertices[0]));
  S2D(segment_get_vertex_attrib(seg0, 1, S2D_POSITION, &seg0_vertices[1]));

  /* Fetch the vertices of the segment 1 */
  S2D(segment_get_vertex_attrib(seg1, 0, S2D_POSITION, &seg1_vertices[0]));
  S2D(segment_get_vertex_attrib(seg1, 1, S2D_POSITION, &seg1_vertices[1]));

  /* Look for the vertex shared by the 2 segments */
  for(seg0_ivertex = 0; seg0_ivertex < 2 && seg0_vert < 0; ++seg0_ivertex) {
  for(seg1_ivertex = 0; seg1_ivertex < 2 && seg1_vert < 0; ++seg1_ivertex) {
    const int vertex_eq = f2_eq_eps
      (seg0_vertices[seg0_ivertex].value,
       seg1_vertices[seg1_ivertex].value,
       1.e-6f);
    if(vertex_eq) {
      seg0_vert = seg0_ivertex;
      seg1_vert = seg1_ivertex;
      /* We assume that the segments are not degenerated. As a consequence we
       * can break here since a vertex of the segment 0 can be equal to at most
       * one vertex of the segment 1 */
      break;
    }
  }}

  /* The segments do not have a common vertex */
  if(seg0_vert < 0) return 0;

  /* Compute the dirctions from shared vertex to the opposite segment vertex */
  f2_sub(d0, seg0_vertices[(seg0_vert+1)%2].value, seg0_vertices[seg0_vert].value);
  f2_sub(d1, seg1_vertices[(seg1_vert+1)%2].value, seg1_vertices[seg1_vert].value);

  /* Compute the cosine between the segments */
  seg0_len = f2_normalize(d0, d0);
  seg1_len = f2_normalize(d1, d1);
  cos_normals = f2_dot(d0, d1);

  /* The angle formed by the 2 segments is sharp. Do not filter the hit */
  if(cos_normals > SHARP_ANGLE_COS_THRESOLD) return 0;

  /* Compute the length from pos<0|1> to shared vertex */
  f2_sub(d0, seg0_vertices[seg0_vert].value, pos0);
  f2_sub(d1, seg1_vertices[seg1_vert].value, pos1);
  tmp0_len = f2_len(d0);
  tmp1_len = f2_len(d1);

  return (eq_epsf(seg0_len, 0, 1.e-6f) || tmp0_len/seg0_len < ON_VERTEX_EPSILON)
      && (eq_epsf(seg1_len, 0, 1.e-6f) || tmp1_len/seg1_len < ON_VERTEX_EPSILON);
}

#else  /* DIM == 3 */
#define ON_EDGE_EPSILON 1.e-4f
/* Check that `hit' roughly lies on an edge. */
static INLINE int
hit_on_edge
  (const struct s3d_hit* hit,
   const float org[3],
   const float dir[3])
{
  struct s3d_attrib v0, v1, v2;
  float E0[3], E1[3], N[3];
  float tri_2area;
  float hit_2area0;
  float hit_2area1;
  float hit_2area2;
  float hit_pos[3];
  ASSERT(hit && !S3D_HIT_NONE(hit) && org && dir);

  /* Retrieve the triangle vertices */
  S3D(triangle_get_vertex_attrib(&hit->prim, 0, S3D_POSITION, &v0));
  S3D(triangle_get_vertex_attrib(&hit->prim, 1, S3D_POSITION, &v1));
  S3D(triangle_get_vertex_attrib(&hit->prim, 2, S3D_POSITION, &v2));

  /* Compute the triangle area * 2 */
  f3_sub(E0, v1.value, v0.value);
  f3_sub(E1, v2.value, v0.value);
  tri_2area = f3_len(f3_cross(N, E0, E1));

  /* Compute the hit position */
  f3_add(hit_pos, org, f3_mulf(hit_pos, dir, hit->distance));

  /* Compute areas */
  f3_sub(E0, v0.value, hit_pos);
  f3_sub(E1, v1.value, hit_pos);
  hit_2area0 = f3_len(f3_cross(N, E0, E1));
  f3_sub(E0, v1.value, hit_pos);
  f3_sub(E1, v2.value, hit_pos);
  hit_2area1 = f3_len(f3_cross(N, E0, E1));
  f3_sub(E0, v2.value, hit_pos);
  f3_sub(E1, v0.value, hit_pos);
  hit_2area2 = f3_len(f3_cross(N, E0, E1));

  if(hit_2area0 / tri_2area < ON_EDGE_EPSILON
  || hit_2area1 / tri_2area < ON_EDGE_EPSILON
  || hit_2area2 / tri_2area < ON_EDGE_EPSILON)
    return 1;

  return 0;
}

static int
hit_shared_edge
  (const struct s3d_primitive* tri0,
   const struct s3d_primitive* tri1,
   const float pos0[3], /* Tested position onto the triangle 0 */
   const float pos1[3]) /* Tested Position onto the triangle 1 */
{
  struct s3d_attrib tri0_vertices[3]; /* Vertex positions of the triangle 0 */
  struct s3d_attrib tri1_vertices[3]; /* Vertex positions of the triangle 1 */
  float E0[3], E1[3]; /* Temporary variables storing triangle edges */
  float N0[3], N1[3]; /* Temporary Normals */
  float tri0_2area, tri1_2area; /* 2*area of the submitted triangles */
  float tmp0_2area, tmp1_2area;
  float cos_normals;
  int tri0_edge[2] = {-1, -1}; /* Shared edge vertex ids for the triangle 0 */
  int tri1_edge[2] = {-1, -1}; /* Shared edge vertex ids for the triangle 1 */
  int edge_ivertex = 0; /* Temporary variable */
  int tri0_ivertex, tri1_ivertex;
  int iv0, iv1, iv2;
  ASSERT(tri0 && tri1 && pos0 && pos1);

  /* Fetch the vertices of the triangle 0 */
  S3D(triangle_get_vertex_attrib(tri0, 0, S3D_POSITION, &tri0_vertices[0]));
  S3D(triangle_get_vertex_attrib(tri0, 1, S3D_POSITION, &tri0_vertices[1]));
  S3D(triangle_get_vertex_attrib(tri0, 2, S3D_POSITION, &tri0_vertices[2]));

  /* Fetch the vertices of the triangle 1 */
  S3D(triangle_get_vertex_attrib(tri1, 0, S3D_POSITION, &tri1_vertices[0]));
  S3D(triangle_get_vertex_attrib(tri1, 1, S3D_POSITION, &tri1_vertices[1]));
  S3D(triangle_get_vertex_attrib(tri1, 2, S3D_POSITION, &tri1_vertices[2]));

  /* Look for the vertices shared by the 2 triangles */
  for(tri0_ivertex=0; tri0_ivertex < 3 && edge_ivertex < 2; ++tri0_ivertex) {
  for(tri1_ivertex=0; tri1_ivertex < 3 && edge_ivertex < 2; ++tri1_ivertex) {
    const int vertex_eq = f3_eq_eps
      (tri0_vertices[tri0_ivertex].value,
       tri1_vertices[tri1_ivertex].value,
       1.e-6f);
    if(vertex_eq) {
      tri0_edge[edge_ivertex] = tri0_ivertex;
      tri1_edge[edge_ivertex] = tri1_ivertex;
      ++edge_ivertex;
      /* We assume that the triangles are not degenerated. As a consequence we
       * can break here since a vertex of the triangle 0 can be equal to at
       * most one vertex of the triangle 1 */
      break;
    }
  }}

  /* The triangles do not have a common edge */
  if(edge_ivertex < 2) return 0;

  /* Ensure that the vertices of the shared edge are registered in the right
   * order regarding the triangle vertices, i.e. (0,1), (1,2) or (2,0) */
  if((tri0_edge[0]+1)%3 != tri0_edge[1]) SWAP(int, tri0_edge[0], tri0_edge[1]);
  if((tri1_edge[0]+1)%3 != tri1_edge[1]) SWAP(int, tri1_edge[0], tri1_edge[1]);

  /* Compute the shared edge normal lying in the triangle 0 plane */
  iv0 =  tri0_edge[0];
  iv1 =  tri0_edge[1];
  iv2 = (tri0_edge[1]+1) % 3;
  f3_sub(E0, tri0_vertices[iv1].value, tri0_vertices[iv0].value);
  f3_sub(E1, tri0_vertices[iv2].value, tri0_vertices[iv0].value);
  f3_cross(N0, E0, E1); /* Triangle 0 normal */
  tri0_2area = f3_len(N0);
  f3_cross(N0, N0, E0);

  /* Compute the shared edge normal lying in the triangle 1 plane */
  iv0 =  tri1_edge[0];
  iv1 =  tri1_edge[1];
  iv2 = (tri1_edge[1]+1) % 3;
  f3_sub(E0, tri1_vertices[iv1].value, tri1_vertices[iv0].value);
  f3_sub(E1, tri1_vertices[iv2].value, tri1_vertices[iv0].value);
  f3_cross(N1, E0, E1);
  tri1_2area = f3_len(N1);
  f3_cross(N1, N1, E0);

  /* Compute the cosine between the 2 edge normals */
  f3_normalize(N0, N0);
  f3_normalize(N1, N1);
  cos_normals = f3_dot(N0, N1);

  /* The angle formed by the 2 triangles is sharp */
  if(cos_normals > SHARP_ANGLE_COS_THRESOLD) return 0;

  /* Compute the 2 times the area of the (pos0, shared_edge.vertex0,
   * shared_edge.vertex1) triangles */
  f3_sub(E0, tri0_vertices[tri0_edge[0]].value, pos0);
  f3_sub(E1, tri0_vertices[tri0_edge[1]].value, pos0);
  tmp0_2area = f3_len(f3_cross(N0, E0, E1));

  /* Compute the 2 times the area of the (pos1, shared_edge.vertex0,
   * shared_edge.vertex1) triangles */
  f3_sub(E0, tri1_vertices[tri1_edge[0]].value, pos1);
  f3_sub(E1, tri1_vertices[tri1_edge[1]].value, pos1);
  tmp1_2area = f3_len(f3_cross(N1, E0, E1));

  return (eq_epsf(tri0_2area, 0, 1.e-6f) || tmp0_2area/tri0_2area < ON_EDGE_EPSILON)
      && (eq_epsf(tri1_2area, 0, 1.e-6f) || tmp1_2area/tri1_2area < ON_EDGE_EPSILON);
}
#undef ON_EDGE_EPSILON
#endif /* DIM == 2 */

/* Avoid self-intersection for a ray starting from a planar primitive, i.e. a
 * triangle or a line segment */
static int
XD(hit_filter_function)
  (const struct sXd(hit)* hit,
   const float org[DIM],
   const float dir[DIM],
   void* ray_data,
   void* global_data)
{
  const struct hit_filter_data* filter_data = ray_data;
  const struct sXd(hit)* hit_from = &filter_data->XD(hit);
  (void)org, (void)dir, (void)global_data;

  if(!ray_data || SXD_HIT_NONE(hit_from)) return 0; /* No filtering */

  if(SXD_PRIMITIVE_EQ(&hit_from->prim, &hit->prim)) return 1;

  if(eq_epsf(hit->distance, 0, (float)filter_data->epsilon)) {
    float pos[DIM];
    fX(add)(pos, org, fX(mulf)(pos, dir, hit->distance));
    /* If the targeted point is near of the origin, check that it lies on an
     * edge/vertex shared by the 2 primitives */
#if DIM == 2
    return hit_shared_vertex(&hit_from->prim, &hit->prim, org, pos);
#else
    return hit_shared_edge(&hit_from->prim, &hit->prim, org, pos);
#endif
  }
  return 0;
}

/* Retrieve the indices of `struct geometry' primitive */
static void
XD(geometry_indices)(const unsigned iprim, unsigned out_ids[DIM], void* data)
{
  struct geometry* ctx = data;
  size_t ids[DIM];
  int i;
  ASSERT(ctx && out_ids);
  ctx->indices(iprim, ids, ctx->data);
  FOR_EACH(i, 0, DIM) out_ids[i] = (unsigned)ids[i];
}

/* Retrieve the coordinates of `struct geometry' vertex */
static void
XD(geometry_position)(const unsigned ivert, double out_pos[DIM], void* data)
{
  struct geometry* ctx = data;
  double pos[DIM];
  int i;
  ASSERT(ctx && out_pos);
  ctx->position(ivert, pos, ctx->data);
  FOR_EACH(i, 0, DIM) out_pos[i] = pos[i];
}

/* Retrieve the indices of a primitive of a Star-EncXD descriptor */
static void
XD(scene_indices)(const unsigned iprim, unsigned ids[DIM], void* data)
{
  struct sencXd(scene)* scn = data;
  SENCXD(scene_get_primitive(scn, iprim, ids));
}

/* Retrieve the coordinates of a vertex of a Star-EncXD descriptor */
static void
XD(scene_position)(const unsigned ivert, float out_pos[DIM], void* data)
{
  struct sencXd(scene)* scn = data;
  double pos[3];
  int i;
  SENCXD(scene_get_vertex(scn, ivert, pos));
  FOR_EACH(i, 0, DIM) out_pos[i] = (float)pos[i];
}

/* Retrieve the indices of a primitive of a Star-EncXD enclosure */
static void
XD(enclosure_indices)(const unsigned iprim, unsigned ids[DIM], void* data)
{
  struct sencXd(enclosure)* enc = data;
  SENCXD(enclosure_get_primitive(enc, iprim, ids));
}

/* Retrieve the coordinates of a vertex of a Star-EncXD encolsure */
static void
XD(enclosure_position)(const unsigned ivert, float out_pos[DIM], void* data)
{
  struct sencXd(enclosure)* enc = data;
  double pos[DIM];
  int i;
  ASSERT(out_pos);
  SENCXD(enclosure_get_vertex(enc, ivert, pos));
  FOR_EACH(i, 0, DIM) out_pos[i] = (float)pos[i];
}

/* Use Star-EncXD to analyze the user defined data. It essentially cleans-up
 * the geometry and extracts the enclosures wrt to the submitted media. Note
 * that data inconsistencies are also detected by this function. In this case
 * an error is returned. */
static res_T
XD(run_analyze)
  (struct sdis_scene* scn,
   const size_t nprims, /* #primitives */
   sdis_get_primitive_indices_T indices,
   sdis_get_primitive_interface_T interf,
   const size_t nverts, /* #vertices */
   sdis_get_vertex_position_T position,
   void* ctx,
   struct sencXd(scene)** out_scn)
{
  struct geometry geom;
  struct sencXd(device)* senc = NULL;
  struct sencXd(scene)* senc3d_scn = NULL;
  unsigned count;
  res_T res = RES_OK;
  ASSERT(scn && nprims && indices && interf && nverts && position && out_scn);

  res = sencXd(device_create)(scn->dev->logger, scn->dev->allocator,
    scn->dev->nthreads, scn->dev->verbose, &senc);
  if(res != RES_OK) goto error;

  /* Setup the geometry data */
  geom.indices = indices;
  geom.interf = interf;
  geom.position = position;
  geom.data = ctx;
  res = sencXd(scene_create)(senc,
    SENCXD_(CONVENTION_NORMAL_BACK) | SENCXD_(CONVENTION_NORMAL_OUTSIDE),
    (unsigned)nprims, XD(geometry_indices), geometry_media,
    (unsigned)nverts, XD(geometry_position), &geom, &senc3d_scn);
  if(res != RES_OK) goto error;
  /* With il-formed scenes, scene creation can success without being able
   * to extract enclosures; in this case just fail */
  res = sencXd(scene_get_enclosure_count(senc3d_scn, &count));
  if(res != RES_OK) goto error;

exit:
  if(senc) SENCXD(device_ref_put(senc));
  if(out_scn) *out_scn = senc3d_scn;
  return res;
error:
  if(senc3d_scn) {
    SENCXD(scene_ref_put(senc3d_scn));
    senc3d_scn = NULL;
  }
  goto exit;
}

/* Register the media and the interfaces, map each primitive to its interface
 * and associated to each primitive the identifier of the enclosures that it
 * splits. */
static res_T
XD(setup_properties)
  (struct sdis_scene* scn,
   struct sencXd(scene)* senc3d_scn,
   void (*interf)(const size_t itri, struct sdis_interface**, void*),
   void* ctx)
{
  unsigned iprim, nprims;
  res_T res = RES_OK;
  ASSERT(scn && senc3d_scn && interf);

  clear_properties(scn);

  SENCXD(scene_get_primitives_count(senc3d_scn, &nprims));
  FOR_EACH(iprim, 0, nprims) {
    struct prim_prop* prim_prop;
    struct sdis_interface* itface;
    unsigned enclosures[2];
    unsigned id;
    int i;
    double* enc_upper_bound;
    size_t ninterfaces;

    /* Fetch the enclosures that the segment/triangle splits */
    SENCXD(scene_get_primitive_enclosures(senc3d_scn, iprim, enclosures));

    /* Fetch the interface of the primitive */
    interf(iprim, &itface, ctx);

    /* Check that the interface is already registered against the scene */
    id = interface_get_id(itface);
    ninterfaces = darray_interf_size_get(&scn->interfaces);
    if(id >= ninterfaces) {
      res = darray_interf_resize(&scn->interfaces, id + 1);
      if(res != RES_OK) goto error;
    }
    if(darray_interf_cdata_get(&scn->interfaces)[id]) {
      ASSERT(darray_interf_cdata_get(&scn->interfaces)[id] == itface);
    } else {
      SDIS(interface_ref_get(itface));
      darray_interf_data_get(&scn->interfaces)[id] = itface;
    }

    /* Register the interface media against the scene */
    res = register_medium(scn, itface->medium_front);
    if(res != RES_OK) goto error;
    res = register_medium(scn, itface->medium_back);
    if(res != RES_OK) goto error;

    /* Allocate primitive properties */
    res = darray_prim_prop_resize(&scn->prim_props, iprim+1);
    if(res != RES_OK) goto error;

    /* Setup primitive properties */
    prim_prop = darray_prim_prop_data_get(&scn->prim_props) + iprim;
    prim_prop->interf = itface;
    prim_prop->front_enclosure = enclosures[0];
    prim_prop->back_enclosure = enclosures[1];

    /* Build per-interface hc upper bounds in a tmp table */
    FOR_EACH(i, 0, 2) {
      double hc_ub = interface_get_convection_coef_upper_bound(itface);
      enc_upper_bound = htable_d_find(&scn->tmp_hc_ub, enclosures+i);
      if(!enc_upper_bound) {
        res = htable_d_set(&scn->tmp_hc_ub, enclosures+i, &hc_ub);
      } else {
        *enc_upper_bound = MMAX(*enc_upper_bound, hc_ub);
      }
    }
  }

exit:
  return res;
error:
  clear_properties(scn);
  goto exit;
}

/* Build the Star-XD scene view of the whole scene */
static res_T
XD(setup_scene_geometry)(struct sdis_scene* scn, struct sencXd(scene)* senc3d_scn)
{
  struct sXd(device)* sXd_dev = NULL;
  struct sXd(shape)* sXd_shape = NULL;
  struct sXd(scene)* sXd_scn = NULL;
  struct sXd(vertex_data) vdata = SXD_VERTEX_DATA_NULL;
  unsigned nprims, nverts;
  res_T res = RES_OK;
  ASSERT(scn && senc3d_scn);

  SENCXD(scene_get_vertices_count(senc3d_scn, &nverts));

  /* Setup the vertex data */
  vdata.usage = SXD_POSITION;
  vdata.type = SXD_FLOATX;
  vdata.get = XD(scene_position);

  /* Create the Star-XD geometry of the whole scene */
  #define CALL(Func)  { if(RES_OK != (res = Func)) goto error; } (void)0
  sXd_dev = scn->dev->sXd(dev);
  SENCXD(scene_get_primitives_count(senc3d_scn, &nprims));
#if DIM == 2
  CALL(sXd(shape_create_line_segments)(sXd_dev, &sXd_shape));
  CALL(sXd(line_segments_set_hit_filter_function)(sXd_shape,
    XD(hit_filter_function), NULL));
  CALL(sXd(line_segments_setup_indexed_vertices)(sXd_shape, nprims,
    XD(scene_indices), nverts, &vdata, 1, senc3d_scn));
#else
  CALL(sXd(shape_create_mesh)(sXd_dev, &sXd_shape));
  CALL(sXd(mesh_set_hit_filter_function)(sXd_shape, XD(hit_filter_function), NULL));
  CALL(sXd(mesh_setup_indexed_vertices)(sXd_shape, nprims, XD(scene_indices),
    nverts, &vdata, 1, senc3d_scn));
#endif
  CALL(sXd(scene_create)(sXd_dev, &sXd_scn));
  CALL(sXd(scene_attach_shape)(sXd_scn, sXd_shape));
  CALL(sXd(scene_view_create)(sXd_scn, SXD_TRACE|SXD_GET_PRIMITIVE,
    &scn->sXd(view)));
  #undef CALL

exit:
  if(sXd_shape) SXD(shape_ref_put(sXd_shape));
  if(sXd_scn) SXD(scene_ref_put(sXd_scn));
  return res;
error:
  if(scn->sXd(view)) SXD(scene_view_ref_put(scn->sXd(view)));
  goto exit;
}

/* Build the Star-XD scene view of a specific enclosure and map their local
 * primitive id to their primitive id in the whole scene */
static res_T
XD(setup_enclosure_geometry)(struct sdis_scene* scn, struct sencXd(enclosure)* enc)
{
  struct sXd(device)* sXd_dev = NULL;
  struct sXd(scene)* sXd_scn = NULL;
  struct sXd(shape)* sXd_shape = NULL;
  struct sXd(vertex_data) vdata = SXD_VERTEX_DATA_NULL;
  struct enclosure enc_dummy;
  struct enclosure* enc_data;
  float S, V;
  double* p_ub;
  unsigned iprim, nprims, nverts;
  struct sencXd(enclosure_header) header;
  res_T res = RES_OK;
  ASSERT(scn && enc);

  enclosure_init(scn->dev->allocator, &enc_dummy);

  SENCXD(enclosure_get_header(enc, &header));
  sXd_dev = scn->dev->sXd(dev);
  nprims = header.primitives_count;
  nverts = header.vertices_count;

  /* Register the enclosure into the scene. Use a dummy data on their
   * registration; in order to avoid a costly copy, we are going to setup the
   * registered data rather than a local data that would be then registered. In
   * other words, the following hash table registration can be seen as an
   * allocation of the enclosure data to setup. */
  res = htable_enclosure_set(&scn->enclosures, &header.enclosure_id, &enc_dummy);
  if(res != RES_OK) goto error;

  /* Fetch the data of the registered enclosure */
  enc_data = htable_enclosure_find(&scn->enclosures, &header.enclosure_id);
  ASSERT(enc_data != NULL);

    /* Setup the vertex data */
  vdata.usage = SXD_POSITION;
  vdata.type = SXD_FLOATX;
  vdata.get = XD(enclosure_position);

  /* Create the Star-XD geometry */
  #define CALL(Func)  { if(RES_OK != (res = Func)) goto error; } (void)0
#if DIM == 2
  CALL(sXd(shape_create_line_segments)(sXd_dev, &sXd_shape));
  CALL(sXd(line_segments_setup_indexed_vertices)(sXd_shape, nprims,
    XD(enclosure_indices), nverts, &vdata, 1, enc));
#else
  CALL(sXd(shape_create_mesh)(sXd_dev, &sXd_shape));
  CALL(sXd(mesh_setup_indexed_vertices)(sXd_shape, nprims, XD(enclosure_indices),
    nverts, &vdata, 1, enc));
#endif
  CALL(sXd(scene_create)(sXd_dev, &sXd_scn));
  CALL(sXd(scene_attach_shape)(sXd_scn, sXd_shape));
  CALL(sXd(scene_view_create)(sXd_scn, SXD_SAMPLE|SXD_TRACE, &enc_data->sXd(view)));

  /* Compute the S/V ratio */
#if DIM == 2
  CALL(s2d_scene_view_compute_contour_length(enc_data->s2d_view, &S));
  CALL(s2d_scene_view_compute_area(enc_data->s2d_view, &V));
#else
  CALL(s3d_scene_view_compute_area(enc_data->s3d_view, &S));
  CALL(s3d_scene_view_compute_volume(enc_data->s3d_view, &V));
#endif
  enc_data->V = V;
  enc_data->S_over_V = S / V;
  ASSERT(enc_data->S_over_V >= 0);
  #undef CALL

    /* Set enclosure hc upper bound regardless of its media being a fluid */
  p_ub = htable_d_find(&scn->tmp_hc_ub, &header.enclosure_id);
  ASSERT(p_ub);
  enc_data->hc_upper_bound = *p_ub;

  /* Define the identifier of the enclosure primitives in the whole scene */
  res = darray_uint_resize(&enc_data->local2global, nprims);
  if(res != RES_OK) goto error;
  FOR_EACH(iprim, 0, nprims) {
    enum sencXd(side) side;
    SENCXD(enclosure_get_primitive_id
      (enc, iprim, darray_uint_data_get(&enc_data->local2global)+iprim, &side));
  }

  /* Setup the medium id of the enclosure */
  SENCXD(enclosure_get_medium(enc, 0, &enc_data->medium_id));

exit:
  enclosure_release(&enc_dummy);
  if(sXd_shape) SXD(shape_ref_put(sXd_shape));
  if(sXd_scn) SXD(scene_ref_put(sXd_scn));
  return res;
error:
  htable_enclosure_erase(&scn->enclosures, &header.enclosure_id);
  goto exit;
}

/* Build the Star-XD scene view and define its associated data of the finite
 * fluid enclosures */
static res_T
XD(setup_enclosures)(struct sdis_scene* scn, struct sencXd(scene)* senc3d_scn)
{
  struct sencXd(enclosure)* enc = NULL;
  unsigned ienc, nencs;
  int inner_multi = 0;
  res_T res = RES_OK;
  ASSERT(scn && senc3d_scn);

  SENCXD(scene_get_enclosure_count(senc3d_scn, &nencs));
  FOR_EACH(ienc, 0, nencs) {
    struct sencXd(enclosure_header) header;

    SENCXD(scene_get_enclosure(senc3d_scn, ienc, &enc));
    SENCXD(enclosure_get_header(enc, &header));

    if(header.is_infinite) {
      ASSERT(scn->outer_enclosure_id == UINT_MAX); /* Not set yet */
      scn->outer_enclosure_id = ienc;
    }

    if(header.enclosed_media_count != 1 && !header.is_infinite)
      inner_multi++;

    /* Silently discard infinite enclosures */
    if(!header.is_infinite) {
      res = XD(setup_enclosure_geometry)(scn, enc);
      if(res != RES_OK) goto error;
    }
    SENCXD(enclosure_ref_put(enc));
    enc = NULL;
  }

  if(inner_multi) {
    log_info(scn->dev,
      "# Found %d internal enclosure(s) with more than 1 medium.\n",
      inner_multi);
  }

  /* tmp table no more useful */
  htable_d_purge(&scn->tmp_hc_ub);
exit:
  if(enc) SENCXD(enclosure_ref_put(enc));
  return res;
error:
  goto exit;
}

/* Create a Stardis scene */
static res_T
XD(scene_create)
  (struct sdis_device* dev,
   const struct sdis_scene_create_args* args,
   struct sdis_scene** out_scn)
{
  struct sencXd(scene)* senc3d_scn = NULL;
  struct sdis_scene* scn = NULL;
  res_T res = RES_OK;

  if(!dev || !check_sdis_scene_create_args(args) || !out_scn) {
    res = RES_BAD_ARG;
    goto error;
  }

  scn = MEM_CALLOC(dev->allocator, 1, sizeof(struct sdis_scene));
  if(!scn) {
    log_err(dev, "%s: could not allocate the Stardis scene.\n", FUNC_NAME);
    res = RES_MEM_ERR;
    goto error;
  }

  ref_init(&scn->ref);
  SDIS(device_ref_get(dev));
  scn->dev = dev;
  scn->fp_to_meter = args->fp_to_meter;
  scn->ambient_radiative_temperature = args->trad;
  scn->reference_temperature = args->tref;
  scn->outer_enclosure_id = UINT_MAX;
  darray_interf_init(dev->allocator, &scn->interfaces);
  darray_medium_init(dev->allocator, &scn->media);
  darray_prim_prop_init(dev->allocator, &scn->prim_props);
  htable_enclosure_init(dev->allocator, &scn->enclosures);
  htable_d_init(dev->allocator, &scn->tmp_hc_ub);

  res = XD(run_analyze)
    (scn,
     args->nprimitives,
     args->get_indices,
     args->get_interface,
     args->nvertices,
     args->get_position,
     args->context,
     &senc3d_scn);
  if(res != RES_OK) {
    log_err(dev, "%s: error during the scene analysis.\n", FUNC_NAME);
    goto error;
  }
  res = XD(setup_properties)(scn, senc3d_scn, args->get_interface, args->context);
  if(res != RES_OK) {
    log_err(dev, "%s: could not setup the scene interfaces and their media.\n",
      FUNC_NAME);
    goto error;
  }
  res = XD(setup_scene_geometry)(scn, senc3d_scn);
  if(res != RES_OK) {
    log_err(dev, "%s: could not setup the scene geometry.\n", FUNC_NAME);
    goto error;
  }
  res = XD(setup_enclosures)(scn, senc3d_scn);
  if(res != RES_OK) {
    log_err(dev, "%s: could not setup the enclosures.\n", FUNC_NAME);
    goto error;
  }
  scn->sencXd(scn) = senc3d_scn;

exit:
  if(out_scn) *out_scn = scn;
  return res;
error:
  if(senc3d_scn) SENCXD(scene_ref_put(senc3d_scn));
  if(scn) {
    SDIS(scene_ref_put(scn));
    scn = NULL;
  }
  goto exit;
}

static res_T
XD(scene_find_closest_point)
  (const struct sdis_scene* scn,
   const double pos[3],
   const double radius,
   size_t* iprim,
   double uv[2])
{
  struct sXd(hit) hit;
  float query_pos[DIM];
  float query_radius;
  res_T res = RES_OK;

  if(!scn || !pos || radius <= 0 || !iprim || !uv
  || scene_is_2d(scn) != (DIM == 2)) {
    res = RES_BAD_ARG;
    goto error;
  }

  /* Avoid a null query radius due to casting in single-precision */
  query_radius = MMAX((float)radius, FLT_MIN);

  fX_set_dX(query_pos, pos);
  res = sXd(scene_view_closest_point)
    (scn->sXd(view), query_pos, query_radius, NULL, &hit);
  if(res != RES_OK) {
#if DIM == 2
    log_err(scn->dev,
      "%s: error querying the closest position at {%g, %g} "
      "for a radius of %g -- %s.\n",
      FUNC_NAME, SPLIT2(query_pos), query_radius, res_to_cstr(res));
#else
   log_err(scn->dev,
      "%s: error querying the closest position at {%g, %g, %g} "
      "for a radius of %g -- %s.\n",
      FUNC_NAME, SPLIT3(query_pos), query_radius, res_to_cstr(res));
#endif
   goto error;
  }

  if(SXD_HIT_NONE(&hit)) {
    *iprim = SDIS_PRIMITIVE_NONE;
  } else {
    *iprim = hit.prim.scene_prim_id;
#if DIM == 2
    uv[0] = hit.u;
#else
    uv[0] = hit.uv[0];
    uv[1] = hit.uv[1];
#endif
  }

exit:
  return res;
error:
  goto exit;
}

/*******************************************************************************
 * Local functions
 ******************************************************************************/
static INLINE res_T
XD(scene_get_medium)
  (const struct sdis_scene* scn,
   const double pos[DIM],
   struct get_medium_info* info, /* May be NULL */
   struct sdis_medium** out_medium)
{
  struct sdis_medium* medium = NULL;
  size_t iprim, nprims;
  size_t nfailures = 0;
  const size_t max_failures = 10;
  float P[DIM];
  /* Range of the parametric coordinate into which positions are challenged */
#if DIM == 2
  float st[3];
#else
  float st[3][2];
#endif
  size_t nsteps = 3;
  res_T res = RES_OK;
  ASSERT(scn && pos);

#if DIM == 2
  st[0] = 0.25f;
  st[1] = 0.50f;
  st[2] = 0.75f;
#else
  f2(st[0], 1.f/6.f, 5.f/12.f);
  f2(st[1], 5.f/12.f, 1.f/6.f);
  f2(st[2], 5.f/12.f, 5.f/12.f);
#endif

  fX_set_dX(P, pos);

  SXD(scene_view_primitives_count(scn->sXd(view), &nprims));
  FOR_EACH(iprim, 0, nprims) {
    struct sXd(hit) hit;
    struct sXd(attrib) attr;
    struct sXd(primitive) prim;
    size_t iprim2;
    const float range[2] = {0.f, FLT_MAX};
    float N[DIM], dir[DIM], cos_N_dir;
    size_t istep = 0;

    /* 1 primitive over 2, take a primitive from the end of the primitive list.
     * When primitives are sorted in a coherent manner regarding their
     * position, this strategy avoids to test primitives that are going to be
     * rejected of the same manner due to possible numerical issues of the
     * resulting intersection. */
    if((iprim % 2) == 0) {
      iprim2 = iprim / 2;
    } else {
      iprim2 = nprims - 1 - (iprim / 2);
    }

    do {
      /* Retrieve a position onto the primitive */
      SXD(scene_view_get_primitive(scn->sXd(view), (unsigned)iprim2, &prim));
      SXD(primitive_get_attrib(&prim, SXD_POSITION, st[istep], &attr));

      /* Trace a ray from the random walk vertex toward the retrieved primitive
       * position */
      fX(normalize)(dir, fX(sub)(dir, attr.value, P));
      SXD(scene_view_trace_ray(scn->sXd(view), P, dir, range, NULL, &hit));

      /* Unforeseen error. One has to intersect a primitive ! */
      if(SXD_HIT_NONE(&hit)) {
        ++nfailures;
        if(nfailures < max_failures) {
          continue;
        } else {
          res = RES_BAD_ARG;
          goto error;
        }
      }
    /* Discard the hit if it is on a vertex/edge, and target a new position
     * onto the current primitive */
    } while((SXD_HIT_NONE(&hit) || HIT_ON_BOUNDARY(&hit, P, dir))
         && ++istep < nsteps);

    /* The hits of all targeted positions on the current primitive are on
     * vertices. Challenge positions on another primitive. */
    if(istep >= nsteps) continue;

    fX(normalize)(N, hit.normal);
    cos_N_dir = fX(dot)(N, dir);

    /* Not too close and not roughly orthognonal */
    if(hit.distance > 1.e-6 && absf(cos_N_dir) > 1.e-2f) {
      const struct sdis_interface* interf;
      interf = scene_get_interface(scn, hit.prim.prim_id);
      medium = interface_get_medium
        (interf, cos_N_dir < 0 ? SDIS_FRONT : SDIS_BACK);

      /* Register the get_medium_info */
      if(info) {
        fX(set)(info->pos_tgt, attr.value);
        fX(set)(info->ray_org, P);
        fX(set)(info->ray_dir, dir);
        info->XD(hit) = hit;
      }
      break;
    }
  }

  if(iprim >= nprims) {
    res = RES_BAD_OP;
    goto error;
  }

  if(iprim > 10 && iprim > (size_t)((double)nprims * 0.05)) {
    log_warn(scn->dev,
      "%s: performance issue. Up to %lu primitives were tested to define the "
      "current medium at {%g, %g, %g}.\n",
      FUNC_NAME, (unsigned long)iprim, SPLIT3(P));
  }

exit:
  *out_medium = medium;
  return res;
error:
#if DIM == 2
  log_err(scn->dev, "%s: could not retrieve the medium at {%g, %g}.\n",
    FUNC_NAME, SPLIT2(pos));
#else
  log_err(scn->dev, "%s: could not retrieve the medium at {%g, %g, %g}.\n",
    FUNC_NAME, SPLIT3(pos));
#endif
  goto exit;
}

static INLINE res_T
XD(scene_get_medium_in_closed_boundaries)
  (const struct sdis_scene* scn,
   const double pos[DIM],
   struct sdis_medium** out_medium)
{
  struct sdis_medium* medium = NULL;
  float P[DIM];
  float frame[DIM*DIM];
  float dirs[6][3] = {{1,0,0},{-1,0,0},{0,1,0},{0,-1,0},{0,0,1},{0,0,-1}};
  int idir;
  res_T res = RES_OK;
  ASSERT(scn && pos);

  /* Build a frame that will be used to rotate the main axis by PI/4 around
   * each axis. This can avoid numerical issues when geometry is discretized
   * along the main axis */
#if DIM == 2
  f22_rotation(frame, (float)PI/4);
#else
/*  N[0] = N[1] = N[2] = (float)(1.0 / sqrt(3.0));*/
/*  f33_basis(frame, N);*/
  f33_rotation(frame, (float)PI/4, (float)PI/4, (float)PI/4);
#endif

  fX_set_dX(P, pos);
  FOR_EACH(idir, 0, 2*DIM) {
    struct sXd(hit) hit;
    float N[DIM];
    const float range[2] = {0.f, FLT_MAX};
    float cos_N_dir;

    /* Transform the directions to avoid to be aligned with the axis */
    fXX_mulfX(dirs[idir], frame, dirs[idir]);

    /* Trace a ray from the random walk vertex toward the retrieved primitive
     * position */
    SXD(scene_view_trace_ray(scn->sXd(view), P, dirs[idir], range, NULL, &hit));

    /* Unforeseen error. One has to intersect a primitive ! */
    if(SXD_HIT_NONE(&hit)) continue;

    /* Discard a hits if it lies on an edge/point */
    if(HIT_ON_BOUNDARY(&hit, P, dirs[idir])) continue;

    fX(normalize)(N, hit.normal);
    cos_N_dir = fX(dot)(N, dirs[idir]);

    /* Not too close and not roughly orthogonal */
    if(hit.distance > 1.e-6 && absf(cos_N_dir) > 1.e-2f) {
      const struct sdis_interface* interf;
      interf = scene_get_interface(scn, hit.prim.prim_id);
      medium = interface_get_medium
        (interf, cos_N_dir < 0 ? SDIS_FRONT : SDIS_BACK);
      break;
    }
  }
  if(idir >= 2*DIM) {
    res = XD(scene_get_medium)(scn, pos, NULL, &medium);
    if(res != RES_OK) goto error;
  }

exit:
  *out_medium = medium;
  return res;
error:
#if DIM == 2
  log_err(scn->dev, "%s: could not retrieve the medium at {%g, %g}.\n",
    FUNC_NAME, SPLIT2(pos));
#else
  log_err(scn->dev, "%s: could not retrieve the medium at {%g, %g, %g}.\n",
    FUNC_NAME, SPLIT3(pos));
#endif
  goto exit;
}

#if (SDIS_SCENE_DIMENSION == 2)
#include <star/sencX2d_undefs.h>
#else /* SDIS_SCENE_DIMENSION == 3 */
#include <star/sencX3d_undefs.h>
#endif

#undef SDIS_SCENE_DIMENSION
#undef DIM
#undef sXd
#undef SXD
#undef SXD_VERTEX_DATA_NULL
#undef SXD_POSITION
#undef SXD_FLOAT3
#undef SXD_TRACE
#undef SXD_TRACE
#undef SXD_GET_PRIMITIVE
#undef SXD_HIT_NONE
#undef SXD_PRIMITIVE_EQ
#undef SXD_FLOATX
#undef fX
#undef fX_set_dX
#undef fXX_mulfX
#undef XD
#undef HIT_ON_BOUNDARY

#endif /* !SDIS_SCENE_DIMENSION */
